/*----------------------------------------------------------------------------
 Copyright:      Radig Ulrich  mailto: mail@ulrichradig.de
 Author:         Radig Ulrich
 Remarks:        
 known Problems: none
 Version:        24.10.2007
 Description:    Webserver Applikation
----------------------------------------------------------------------------*/

#include "httpd.h"
#include "webpage.h"
#include "dos/dos.h"


struct http_table http_entry[MAX_TCP_ENTRY];

//Hier wird das codierte Passwort aus config.h gespeichert.
unsigned char http_auth_passwort[20];

//----------------------------------------------------------------------------
//Variablenarry zum einfgen in Webseite %VA@00 bis %VA@09
unsigned int var_array[MAX_VAR_ARRAY] = {10,50,30,0,0,0,0,0,0,0};
//----------------------------------------------------------------------------

PROGMEM char http_header1[]={	"HTTP/1.0 200 Document follows\r\n"
								"Server: AVR_Small_Webserver\r\n"
								"Content-Type: text/html\r\n\r\n"};//LEN 87

PROGMEM char http_header2[]={	"HTTP/1.0 200 Document follows\r\n"
								"Server: AVR_Small_Webserver\r\n"
								"Content-Type: image/jpg\r\n\r\n"};//LEN 87

PROGMEM char http_header3[]={	"HTTP/1.0 401 Unauthorized\r\n"
								"Server: AVR_Small_Webserver\r\n"
								"WWW-Authenticate: Basic realm=\"NeedPassword\""
								"\r\nContent-Type: text/html\r\n\r\n"};//LEN 129

//----------------------------------------------------------------------------
//Kein Zugriff Seite bei keinem Passwort
PROGMEM char Page0[] = {"401 Unauthorized%END"};

unsigned char rx_header_end[5] = {"\r\n\r\n\0"};

//----------------------------------------------------------------------------
//Initialisierung des Httpd Testservers
void httpd_init (void)
{
	//Port fr die schaltbaren Ausgnge
	//Port kann nur bei ENC28J60 benutzt werden 
	//RTL8019 bentigt PORTA
#if USE_ENC28J60
	DDRA = 0x07; 
#endif

	//mal sehen was frei ist
#if USE_RTL8019
#endif

	//HTTP_AUTH_STRING 
	decode_base64((unsigned char*)HTTP_AUTH_STRING,http_auth_passwort);

	//Serverport und Anwendung eintragen
	add_tcp_app (80, (void(*)(unsigned char))httpd);
}
   
//----------------------------------------------------------------------------
//http Testserver
void httpd (unsigned char index)
{
	//Allererste Aufruf des Ports fr diese Anwendung
	//HTTPD_Anwendungsstack lschen
	if(tcp_entry[index].app_status==1)
	{
		httpd_stack_clear(index);
	}
	
	//HTTP wurde bei dieser Verbindung zum ersten mal aufgerufen oder
	//HTML Header Retransmission!
	if (tcp_entry[index].app_status <= 2)
	{	
		httpd_header_check (index);
	}
	
	//Der Header wurde gesendet und mit ACK besttigt (tcp_entry[index].app_status+1)
	//war das HTML Packet fertig, oder mssen weitere Daten gesendet werden, oder Retransmission?
	if (tcp_entry[index].app_status > 2 && tcp_entry[index].app_status < 0xFFFE)
	{
		httpd_data_send (index);
	}
	
	//Verbindung kann geschlossen werden! Alle HTML Daten wurden gesendet TCP Port kann
	//geschlossen werden (tcp_entry[index].app_status >= 0xFFFE)!!
	if (tcp_entry[index].app_status >= 0xFFFE)
	{
		tcp_entry[index].app_status = 0xFFFE;
		tcp_Port_close(index);
		return;
	}
	
	return;
}

//----------------------------------------------------------------------------
//HTTPD_STACK lschen
void httpd_stack_clear (unsigned char index)
{
	http_entry[index].http_header_type =0;
	http_entry[index].first_switch = 0;
	http_entry[index].http_auth = HTTP_AUTH_DEFAULT;
	http_entry[index].new_page_pointer = 0;
	http_entry[index].old_page_pointer = 0;
	http_entry[index].auth_ptr = http_auth_passwort;
	http_entry[index].hdr_end_pointer = rx_header_end;
	HTTP_DEBUG("\r\n**** NEUE HTTP ANFORDERUNG ****\r\n\r\n");	
	return;
}

//----------------------------------------------------------------------------
//Eintreffenden Header vom Client checken
void httpd_header_check (unsigned char index)
{
	//finden der Authorization und das Ende im Header auch ber mehrere Packete hinweg!!	
	if(*http_entry[index].hdr_end_pointer != 0)
	{		
		for(unsigned int a=TCP_DATA_START_VAR;a<(TCP_DATA_END_VAR);a++)
		{	
			HTTP_DEBUG("%c",eth_buffer[a]);
			
			if(!http_entry[index].http_auth) 
			{
				if (eth_buffer[a] != *http_entry[index].auth_ptr++)
				{
					http_entry[index].auth_ptr = http_auth_passwort;
				}
				if(*http_entry[index].auth_ptr == 0) 
				{
					http_entry[index].http_auth = 1;
					HTTP_DEBUG("  <---LOGIN OK!--->\r\n");
				}
			}
			
			if (eth_buffer[a] != *http_entry[index].hdr_end_pointer++)
			{
				http_entry[index].hdr_end_pointer = rx_header_end;
			}
			
			//Das Headerende wird mit (CR+LF+CR+LF) angezeigt!
			if(*http_entry[index].hdr_end_pointer == 0) 
			{
				HTTP_DEBUG("<---HEADER ENDE ERREICHT!--->\r\n");
				
				//Schaltet Ausgnge am AVR je nach Headerinhalt
				//oder Startseite (POST) Schaltdaten am Ende vom Header
				if(http_entry[index].http_auth)
					{
					if (strcasestr((char*)&eth_buffer[a],"LED1=1\0")!=0)
					 { PORTA ^= (1<<PA0);}
					if (strcasestr((char*)&eth_buffer[a],"LED2=2\0")!=0)
					 {PORTA ^= (1<<PA1);}
					if (strcasestr((char*)&eth_buffer[a],"LED3=3\0")!=0)
					 {PORTA ^= (1<<PA2);}
					}
				break;
			}
		}
	}
	
      if (strcasestr((char*)&eth_buffer[TCP_DATA_START_VAR],"GET /\0")!=0)
       {
	//Welche datei wird angefordert? Wird diese in der Flashspeichertabelle gefunden?
	unsigned char page_index = 0;
	
	if (!http_entry[index].new_page_pointer)
	{

/* Geht auch ohne
		for(unsigned int a=TCP_DATA_START_VAR+5;a<(TCP_DATA_END_VAR);a++)
		{
			if (eth_buffer[a] == '\r')
			{
				eth_buffer[a] = '\0';
				break;
			}
		}
*/
	
		while(WEBPAGE_TABLE[page_index].filename)
		{
			if (strcasestr((char*)&eth_buffer[TCP_DATA_START_VAR],WEBPAGE_TABLE[page_index].filename)!=0) 
				{
					http_entry[index].http_header_type = 1;
					HTTP_DEBUG("\r\n\r\nDatei gefunden: ");
					//usart_write_str((char*)WEBPAGE_TABLE[page_index].filename);
					HTTP_DEBUG("<----------------\r\n\r\n");	
					if (strcasestr(WEBPAGE_TABLE[page_index].filename,".jpg")!=0)
					{
						http_entry[index].http_header_type = 1;
					}
					if (strcasestr(WEBPAGE_TABLE[page_index].filename,".gif")!=0)
					{
						http_entry[index].http_header_type = 1;
					}	
					if (strcasestr(WEBPAGE_TABLE[page_index].filename,".htm")!=0)
					{
						http_entry[index].http_header_type = 0;	
					}	
					http_entry[index].new_page_pointer = WEBPAGE_TABLE[page_index].page_pointer;
					break;
				}
			page_index++;
		}
	}

	//Standart INDEX.HTM Seite wenn keine andere gefunden wurde
	if (!http_entry[index].new_page_pointer)
	{
		//Besucher Counter
		var_array[MAX_VAR_ARRAY-1]++;
		
		http_entry[index].new_page_pointer = Page1;
		http_entry[index].http_header_type = 0;
	}	
	
        // Seite auf MMC/SD suchen
	if(FATtype!=0)
	 {
          if (eth_buffer[TCP_DATA_START_VAR+5]!=' ' )
           {
            unsigned char by,i;

            i=0; 
            do
             {
              by=eth_buffer[TCP_DATA_START_VAR+i+5];
              if(by==' ') break;
              http_entry[index].filename[i]=by;
              http_entry[index].filename[i+1]=0;
              i++;
             }while(i<12);

	    if(FindName(http_entry[index].filename)==FULL_MATCH)
	     {
//	      usart_write("File found ");
//	      usart_write_str(http_entry[index].filename);
//	      usart_write("\r\n");

	      http_entry[index].new_page_pointer = NULL; // we have a file on MMC !
              http_entry[index].new_file_pointer =  0;
              http_entry[index].old_file_pointer =  0;
              
              // Filetype rausfinden
              if(strncasecmp_P(&FileName[8],PSTR("GIF"),3)==0) http_entry[index].http_header_type=1;
              if(strncasecmp_P(&FileName[8],PSTR("JPG"),3)==0) http_entry[index].http_header_type=1;
              if(strncasecmp_P(&FileName[8],PSTR("HTM"),3)==0) http_entry[index].http_header_type=0;
	     }
	    else
	     {
//	      usart_write("File not found ");
//	      usart_write_str(http_entry[index].filename);
//	      usart_write("\r\n");
	     } 
	   } 
	 }//if(FATtype!=0)  

      }//if (strcasestr((char*)&eth_buffer[TCP_DATA_START_VAR],"GET /\0")!=0)
      
	//Wurde das Ende vom Header nicht erreicht
	//kommen noch weitere Stcke vom Header!
	if (*http_entry[index].hdr_end_pointer != 0)
	{
		//Der Empfang wird Quitiert und es wird auf weiteres Headerstck gewartet
		tcp_entry[index].status =  ACK_FLAG;
		create_new_tcp_packet(0,index);
		//Warten auf weitere Headerpackete
		tcp_entry[index].app_status = 1;
		return;
	}	
	
	//Wurde das Passwort in den ganzen Headerpacketen gefunden?
	//Wenn nicht dann ausfhren und Passwort anfordern!
	if((!http_entry[index].http_auth) && tcp_entry[index].status&PSH_FLAG)
	{	
		//HTTP_AUTH_Header senden!
		http_entry[index].new_page_pointer = Page0;
		memcpy_P((char*)&eth_buffer[TCP_DATA_START_VAR],http_header3,HTTP_AUTH_HEADER_LEN);
		tcp_entry[index].status =  ACK_FLAG | PSH_FLAG;
		create_new_tcp_packet(129,index);
		return;
	}
	
	tcp_entry[index].app_status = 2;
	//Seiten Header wird gesendet
	if(http_entry[index].http_header_type == 1)
	{
		memcpy_P((char*)&eth_buffer[TCP_DATA_START_VAR],http_header2,HTTP_OK_HEADER_LEN);
	}
	if(http_entry[index].http_header_type == 0)
	{
		memcpy_P((char*)&eth_buffer[TCP_DATA_START_VAR],http_header1,HTTP_OK_HEADER_LEN);
	}
	tcp_entry[index].status =  ACK_FLAG | PSH_FLAG;
	create_new_tcp_packet(87,index);
	return;	
}

//----------------------------------------------------------------------------
//Daten Packete an Client schicken
void httpd_data_send (unsigned char index)
{		
	//Passwort wurde im Header nicht gefunden
	if(!http_entry[index].http_auth)
	{
		http_entry[index].new_page_pointer = Page0;
	}
	
	unsigned int a;
	unsigned char str_len;
	char var_conversion_buffer[10];
	
     if(http_entry[index].new_page_pointer!=NULL) // send from flash
      {
	//kein Packet empfangen Retransmission des alten Packetes
	if (tcp_entry[index].status == 0) 
	{
		http_entry[index].new_page_pointer = http_entry[index].old_page_pointer;
	}
	http_entry[index].old_page_pointer = http_entry[index].new_page_pointer;

	for (a = 0;a<(MTU_SIZE-(TCP_DATA_START)-150);a++)
	{
		unsigned char b;
		b = pgm_read_byte(http_entry[index].new_page_pointer++);
		eth_buffer[TCP_DATA_START + a] = b;
		
		//Mssen Variablen ins Packet eingesetzt werden? ===> %VA@00 bis %VA@09
		if (b == '%')
		{
			if (strncasecmp_P("VA@",http_entry[index].new_page_pointer,3)==0)
			{	
				b = (pgm_read_byte(http_entry[index].new_page_pointer+3)-48)*10;
				b +=(pgm_read_byte(http_entry[index].new_page_pointer+4)-48);	
				itoa (var_array[b],var_conversion_buffer,10);
				str_len = strnlen(var_conversion_buffer,MAX_VAR_ARRAY);
				memmove(&eth_buffer[TCP_DATA_START+a],var_conversion_buffer,str_len);
				a = a + (str_len-1);
				http_entry[index].new_page_pointer=http_entry[index].new_page_pointer+5;
			}
			
			//Ersetzen der Variablen %VI@00 bis %VI@09 durch "EIN" oder "AUS"
			if (strncasecmp_P("VI@",http_entry[index].new_page_pointer,3)==0)
			{
				b = (pgm_read_byte(http_entry[index].new_page_pointer+3)-48)*10;
				b +=(pgm_read_byte(http_entry[index].new_page_pointer+4)-48);	
				if(var_array[b] !=0 )
					memmove(&var_conversion_buffer, "EIN\0", 4);
				else
					memmove(&var_conversion_buffer, "AUS\0", 4);
				str_len = strnlen(var_conversion_buffer,MAX_VAR_ARRAY);
				memmove(&eth_buffer[TCP_DATA_START+a],var_conversion_buffer,str_len);
				a += str_len-1;
				http_entry[index].new_page_pointer = http_entry[index].new_page_pointer+5;
			}
			
			//Einsetzen der Variablen %VR@00x bis %VR@09x durch "checked" wenn var_array[00..]=x
			if (strncasecmp_P("VR@",http_entry[index].new_page_pointer,3)==0)
			{
				b = (pgm_read_byte(http_entry[index].new_page_pointer+3)-48)*10;
				b +=(pgm_read_byte(http_entry[index].new_page_pointer+4)-48);	

				if(var_array[b] == (pgm_read_byte(http_entry[index].new_page_pointer+5)-48))
					memmove(&var_conversion_buffer, "checked\0", 8);
				else
					memmove(&var_conversion_buffer, "\0", 1);
						
				str_len = strnlen(var_conversion_buffer,MAX_VAR_ARRAY);
				memmove(&eth_buffer[TCP_DATA_START+a],var_conversion_buffer,str_len);
				a += str_len-1;
				http_entry[index].new_page_pointer = http_entry[index].new_page_pointer+6;
			}
		}
		//wurde das Ende des Packetes erreicht?
		//Verbindung TCP Port kann beim nchsten ACK geschlossen werden
		//Schleife wird abgebrochen keine Daten mehr!!
		if (strncasecmp_P("%END",http_entry[index].new_page_pointer,4)==0)
		{	
			tcp_entry[index].app_status = 0xFFFD;
			a++;
			break;
		}
	}

       }
      else //if(http_entry[index].new_page_pointer!=NULL) // Sende von MMC/SD
       {
        a = 0;
	unsigned char str_len;
	char var_conversion_buffer[10];

//      usart_write("File from MMC\r\n");

	//kein Packet empfangen Retransmission des alten Packetes
	if (tcp_entry[index].status == 0) 
	 {
	  http_entry[index].new_file_pointer = http_entry[index].old_file_pointer;
	 }
	http_entry[index].old_file_pointer = http_entry[index].new_file_pointer;


        if(http_entry[index].http_header_type == 0) // this is a HTML file
         {
          if(Fopen(http_entry[index].filename,'r')==F_OK)
           {
            Fseek(http_entry[index].new_file_pointer, SEEK_SET);

            // Lese die Datei zeichenweise bis Buffer voll, oder Sonderzeichen
            // gefunden wurden wo Variablen eingefgt werden mssen
            for (a = 0;a<(MTU_SIZE-(TCP_DATA_START)-150);a++)
	      {
       		unsigned char b,c,va_found;
       		char va_buffer[5];

       		if(Fread(&c,1)==0) // End of file reached ?
       		 {
       		  tcp_entry[index].app_status = 0xFFFD;
       		  break;
       		 }
               
               if(c=='%') // Beginn von %VA@, %VI@, %VR@?
                {
                 if( (b=Fread((unsigned char *)va_buffer,5)) != 5) // EOF ?
                  {
                   eth_buffer[TCP_DATA_START + a] = c; // '%' in Puffer
                   http_entry[index].new_file_pointer ++;
                   Fseek(-b,SEEK_END); // Datei zurckspulen und Schleife weiter ausfhren
                  }
                 else
                  {
                   va_found=0;
                   
                   if(strncasecmp_P(va_buffer,PSTR("VA@"),3)==0)
                    {
                        http_entry[index].new_file_pointer +=5;
			b = (va_buffer[3]-'0')*10;
			b +=(va_buffer[4]-'0');	
			itoa (var_array[b],var_conversion_buffer,10);

			str_len = strnlen(var_conversion_buffer,MAX_VAR_ARRAY);
			memmove(&eth_buffer[TCP_DATA_START+a],var_conversion_buffer,str_len);
			a = a + (str_len-1); // -1 wegen a++ am Ende der Schleife
			va_found=1;
                    }
                   
		   //Ersetzen der Variablen %VI@00 bis %VI@09 durch "EIN" oder "AUS"
                   if(strncasecmp_P(va_buffer,PSTR("VI@"),3)==0)
		    {
                        http_entry[index].new_file_pointer +=5;
			b = (va_buffer[3]-'0')*10;
			b +=(va_buffer[4]-'0');	
			if(var_array[b] !=0 ) strcpy_P(var_conversion_buffer, PSTR("EIN"));
			else strcpy_P(var_conversion_buffer, PSTR("AUS"));

			str_len = strnlen(var_conversion_buffer,MAX_VAR_ARRAY);
			memmove(&eth_buffer[TCP_DATA_START+a],var_conversion_buffer,str_len);
			a += str_len-1;
			va_found=1;
		    }

  		   //Einsetzen der Variablen %VR@00x bis %VR@09x durch "checked" wenn var_array[00..]=x
                   if(strncasecmp_P(va_buffer,PSTR("VR@"),3)==0)
		    {
                        unsigned char d;
                        
                        http_entry[index].new_file_pointer +=5;
			b = (va_buffer[3]-'0')*10;
			b +=(va_buffer[4]-'0');	

                        Fread(&d,1); // Bse, kein Check auf EOF
                        http_entry[index].new_file_pointer ++; // siehe oben
                        
                	if(var_array[b] == d-'0')
				strcpy_P(var_conversion_buffer, PSTR("checked"));
			else
				var_conversion_buffer[0]=0;

			str_len = strnlen(var_conversion_buffer,MAX_VAR_ARRAY);
			memmove(&eth_buffer[TCP_DATA_START+a],var_conversion_buffer,str_len);
			a += str_len-1;
			va_found=1;
		    }

                   if(va_found==0) // War kein %VA@,%VI@,%VR@. Datei zurckspulen !
                    {
                     eth_buffer[TCP_DATA_START + a] = c; // '%' in Puffer
                     http_entry[index].new_file_pointer ++;
                     Fseek(-b,SEEK_CUR); // Datei zurckspulen und Schleife weiter ausfhren
                    }
                   
                  }
                }
               else //if(c=='%')
                {
                 eth_buffer[TCP_DATA_START + a] = c;
                 http_entry[index].new_file_pointer ++;
       		}
       		
              }//for (a = 0;a<(MTU_SIZE...
            
                    
            Fclose();
           }
          else
           {
            HTTP_DEBUG("File not found !\r\n");
	    tcp_entry[index].app_status = 0xFFFD; // file not found
           } 

         }
         
        if(http_entry[index].http_header_type == 1) // this is a JPEG Picture
         {
          // Don't check for %VA@ or other special features, only send raw data
          // because it is faster
          
          if(Fopen(http_entry[index].filename,'r')==F_OK)
           {
            Fseek(http_entry[index].new_file_pointer, SEEK_SET);

            if( ( a=Fread(&eth_buffer[TCP_DATA_START],MAX_FREAD) ) != MAX_FREAD)
             {
              //HTTP_DEBUG("EOF !\r\n");
	      tcp_entry[index].app_status = 0xFFFD; // EOF reached
             }
                   
            http_entry[index].new_file_pointer += a;
            Fclose();
           }
          else
           {
            HTTP_DEBUG("File not found !\r\n");
	    tcp_entry[index].app_status = 0xFFFD; // file not found
           } 

         }//if(http_entry[index].http_header_type == 1) // this is a JPEG Picture

        }//if(http_entry[index].new_page_pointer!=NULL)

	//Erzeugte Packet kann nun gesendet werden!
	tcp_entry[index].status =  ACK_FLAG | PSH_FLAG;
	create_new_tcp_packet(a,index);
	return;
}


